from ..base import STLSolver
from pydrake.all import MathematicalProgram, ge, le

class DrakeSTLSolver(STLSolver):
    """
    A base class for solvers that use the Drake interface to connect with
    a lower-level solver like Gurobi, Mosek, SNOPT, or IPOPT.
    """
    def __init__(self, spec, sys, x0, T, verbose):
        STLSolver.__init__(self, spec, sys, x0, T, verbose)

        # Create the drake MathematicalProgram instance that will allow
        # us to interface with a MIP solver like Gurobi or Mosek
        self.mp = MathematicalProgram()

        # Create optimization variables
        self.y = self.mp.NewContinuousVariables(self.sys.p, self.T, 'y')
        self.x = self.mp.NewContinuousVariables(self.sys.n, self.T, 'x')
        self.u = self.mp.NewContinuousVariables(self.sys.m, self.T, 'u')
        self.rho = self.mp.NewContinuousVariables(1,'rho')[0]

    def AddRobustnessConstraint(self, rho_min=0.0):
        self.mp.AddConstraint( self.rho >= rho_min )

    def AddRobustnessCost(self):
        self.mp.AddCost(-self.rho)

    def AddControlBounds(self, u_min, u_max):
        for t in range(self.T):
            self.mp.AddConstraint(le(
                self.u[:,t], u_max
            ))
            self.mp.AddConstraint(ge(
                self.u[:,t], u_min
            ))

    def AddStateBounds(self, x_min, x_max):
        for t in range(self.T):
            self.mp.AddConstraint(le(
                self.x[:,t], x_max
            ))
            self.mp.AddConstraint(ge(
                self.x[:,t], x_min
            ))

    def AddQuadraticCost(self, Q, R):
        for t in range(self.T):
            self.mp.AddCost( self.x[:,t].T@Q@self.x[:,t] + self.u[:,t].T@R@self.u[:,t] )



import numpy as np


class NonlinearSystem:
    """
    A class which represents some (possibly nonlinear)
    discrete-time control system

    .. math::

        x_{t+1} = f(x_t, u_t)

        y_t = g(x_t, u_t)

    where

        - :math:`x_t \in \mathbb{R}^n` is a system state,
        - :math:`u_t \in \mathbb{R}^m` is a control input,
        - :math:`y_t \in \mathbb{R}^p` is a system output.

    :param f:   A function representing :math:`f`, which takes two numpy
                arrays (:math:`x_t,u_t`) as input and returns another
                numpy array (:math:`x_{t+1}`).
    :param g:   A function representing :math:`f`, which takes two numpy
                arrays (:math:`x_t,u_t`) as input and returns another
                numpy array (:math:`y_{t}`).
    :param n:   Size of the state vector :math:`x_t`.
    :param m:   Size of the control vector :math:`u_t`.
    :param p:   Size of the output vector :math:`p_t`.

    """
    def __init__(self, f, g, n, m, p):
        # TODO: do some checks on function signature
        self.dynamics_fcn = f
        self.output_fcn = g
        self.n = n
        self.m = m
        self.p = p

    def f(self, x, u):
        """
        Given state :math:`x_t` and control :math:`u_t`, compute
        the forward dynamics

        .. math::

            x_{t+1} = f(x_t, u_t).

        :param x:   The current state :math:`x_t`
        :param u:   The control input :math:`u_t`

        :return:    The subsequent state :math:`x_{t+1}`
        """
        return self.dynamics_fcn(x, u)

    def g(self, x, u):
        """
        Given state :math:`x_t` and control :math:`u_t`, compute
        the output

        .. math::

            y_t = g(x_t, u_t).

        :param x:   The current state :math:`x_t`
        :param u:   The control input :math:`u_t`

        :return:    The output :math:`y_t`
        """
        return self.output_fcn(x,u)

class Unicycle(NonlinearSystem):
    r"""
    A simple nonlinear system representing a 2D mobile robot with
    unicycle dynamics. The robot is controlled by specifing a forward
    velociy :math:`v` and an angular velocity :math:`\omega`. 

    This is example of a non-holonomic system: the robot cannot
    directly control its motion in the horizontal direction.

    The state is given by 

    .. math::

        x = \begin{bmatrix} p_x \\ p_y \\ \theta \end{bmatrix},

    where :math:`p_x` and :math:`p_y` are positions in the plane and
    :math:`\theta` is an orientation. The dynamics are given by

    .. math::

        \dot{x} = \begin{bmatrix} v \cos(\theta) \\ v \sin(\theta) \\ \omega \end{bmatrix}

    and the control input is :math:`u = \begin{bmatrix} v \\ \omega \end{bmatrix}`.
    We use forward Euler integration to transform this into a discrete-time system:

    .. math::

        x_{t+1} = x_t + \dot{x}~dt.

    The system output is simply the state of the robot, :math:`y_t = x_t`.

    :param dt:  Discretization step size (for forward Euler integration)
    """
    def __init__(self, dt):
        self.dt = dt
        
        # State, control, and output sizes
        self.n = 3
        self.m = 2
        self.p = 3

    def f(self, x, u):
        v = u[0]      # linear velocity
        omega = u[1]  # angular velocity
        theta = x[2]  # orientation

        xdot = np.array([ v * np.cos(theta),
                          v * np.sin(theta),
                          omega])

        return x + self.dt * xdot

    def g(self, x, u):
        return x



import numpy as np
from .formula import STLFormula

class NonlinearPredicate(STLFormula):
    """
    A nonlinear STL predicate:math:`\pi` defined by

    .. math::

        g(y_t) \geq 0

    where :math:`y_t \in \mathbb{R}^d` is the value of the signal
    at a given timestep :math:`t`, and :math:`g : \mathbb{R}^d \\to \mathbb{R}`.
    
    :param g:       A function mapping the signal at a given timestep to 
                    a scalar value. 
    :param d:       An integer expressing the dimension of the signal y.
    :param name:    (optional) a string used to identify this predicate.
    """
    def __init__(self, g, d, name=None):
        self.d = d
        self.name = name
        self.g = g

    def negation(self):
        if self.name is None:
            newname = None
        else:
            newname = "not " + self.name

        negative_g = lambda y : -self.g(y)
        return NonlinearPredicate(negative_g, self.d, name=newname)

    def robustness(self, y, t):
        assert isinstance(y, np.ndarray), "y must be a numpy array"
        assert isinstance(t, int), "timestep t must be an integer"
        assert y.shape[0] == self.d, "y must be of shape (d,T)"
        assert y.shape[1] > t, "requested timestep %s, but y only has %s timesteps" % (t, y.shape[1])

        return np.array([self.g(y[:,t])])

    def is_predicate(self):
        return True

    def is_state_formula(self):
        return True

    def is_disjunctive_state_formula(self):
        return True

    def is_conjunctive_state_formula(self):
        return True
    
    def get_all_inequalities(self):
        raise NotImplementedError("linear inequalities are not defined for nonlinear predicates")

    def __str__(self):
        if self.name is None:
            return "{ Nonlinear Predicate }"
        else:
            return "{ Predicate " + self.name + " }"

class LinearPredicate(STLFormula):
    """
    A linear STL predicate :math:`\pi` defined by

    .. math::

        a^Ty_t - b \geq 0

    where :math:`y_t \in \mathbb{R}^d` is the value of the signal
    at a given timestep :math:`t`, :math:`a \in \mathbb{R}^d`,
    and :math:`b \in \mathbb{R}`.

    :param a:       a numpy array or list representing the vector :math:`a`
    :param b:       a list, numpy array, or scalar representing :math:`b`
    :param name:    (optional) a string used to identify this predicate.
    """
    def __init__(self, a, b, name=None):
        # Convert provided constraints to numpy arrays
        self.a = np.asarray(a).reshape((-1,1))
        self.b = np.atleast_1d(b)

        # Some dimension-related sanity checks
        assert (self.a.shape[1] == 1), "a must be of shape (d,1)"
        assert (self.b.shape == (1,)), "b must be of shape (1,)"

        # Store the dimensionality of y_t
        self.d = self.a.shape[0]

        # A unique string describing this predicate
        self.name = name

    def negation(self):
        if self.name is None:
            newname = None
        else:
            newname = "not " + self.name
        return LinearPredicate(-self.a, -self.b, name=newname)

    def robustness(self, y, t):
        assert isinstance(y, np.ndarray), "y must be a numpy array"
        assert isinstance(t, int), "timestep t must be an integer"
        assert y.shape[0] == self.d, "y must be of shape (d,T)"
        assert y.shape[1] > t, "requested timestep %s, but y only has %s timesteps" % (t, y.shape[1])

        return self.a.T@y[:,t] - self.b

    def is_predicate(self):
        return True

    def is_state_formula(self):
        return True

    def is_disjunctive_state_formula(self):
        return True

    def is_conjunctive_state_formula(self):
        return True

    def get_all_inequalities(self):
        A = -self.a.T
        b = -self.b
        return (A,b)

    def __str__(self):
        if self.name is None:
            return "{ Predicate %s*y >= %s }" % (self.a, self.b)
        else:
            return "{ Predicate " + self.name + " }"




from ..base import STLSolver
from ...STL import LinearPredicate, NonlinearPredicate
import numpy as np

import gurobipy as gp
from gurobipy import GRB

import time

class GurobiMICPSolver(STLSolver):
    """
    Given an :class:`.STLFormula` :math:`\\varphi` and a :class:`.LinearSystem`,
    solve the optimization problem

    .. math::

        \min & -\\rho^{\\varphi}(y_0,y_1,\dots,y_T) + \sum_{t=0}^T x_t^TQx_t + u_t^TRu_t

        \\text{s.t. } & x_0 \\text{ fixed}

        & x_{t+1} = A x_t + B u_t

        & y_{t} = C x_t + D u_t

        & \\rho^{\\varphi}(y_0,y_1,\dots,y_T) \geq 0

    with Gurobi using mixed-integer convex programming. This gives a globally optimal
    solution, but may be computationally expensive for long and complex specifications.
    
    .. note::

        This class implements the algorithm described in

        Belta C, et al.
        *Formal methods for control synthesis: an optimization perspective*.
        Anual Review of Control, Robotics, and Autonomous Systems, 2019.

    :param spec:            An :class:`.STLFormula` describing the specification.
    :param sys:             A :class:`.LinearSystem` describing the system dynamics.
    :param x0:              A ``(n,1)`` numpy matrix describing the initial state.
    :param T:               A positive integer fixing the total number of timesteps :math:`T`.
    :param M:               (optional) A large positive scalar used to rewrite ``min`` and ``max`` as
                            mixed-integer constraints. Default is ``1000``.
    :param robustness_cost: (optional) Boolean flag for adding a linear cost to maximize
                            the robustness measure. Default is ``True``.
    :param presolve:        (optional) A boolean indicating whether to use Gurobi's
                            presolve routines. Default is ``True``.
    :param verbose:         (optional) A boolean indicating whether to print detailed
                            solver info. Default is ``True``.
    """

    def __init__(self, spec, sys, x0, T, M=1000, robustness_cost=True, 
            presolve=True, verbose=True):
        assert M > 0, "M should be a (large) positive scalar"
        super().__init__(spec, sys, x0, T, verbose)

        self.M = float(M)
        self.presolve = presolve

        # Set up the optimization problem
        self.model = gp.Model("STL_MICP")

        # Set some model parameters
        if not self.presolve:
            self.model.setParam('Presolve', 0)
        if not self.verbose:
            self.model.setParam('OutputFlag', 0)

        if self.verbose:
            print("Setting up optimization problem...")
            st = time.time()  # for computing setup time

        # Create optimization variables
        self.y = self.model.addMVar((self.sys.p, self.T), lb=-float('inf'), name='y')
        self.x = self.model.addMVar((self.sys.n, self.T), lb=-float('inf'), name='x')
        self.u = self.model.addMVar((self.sys.m, self.T), lb=-float('inf'), name='u')
        self.rho = self.model.addMVar(1,name="rho",lb=0.0) # lb sets minimum robustness

        # Add cost and constraints to the optimization problem
        self.AddDynamicsConstraints()
        self.AddSTLConstraints()
        self.AddRobustnessConstraint()
        if robustness_cost:
            self.AddRobustnessCost()

        if self.verbose:
            print(f"Setup complete in {time.time()-st} seconds.")

    def AddControlBounds(self, u_min, u_max):
        for t in range(self.T):
            self.model.addConstr( u_min <= self.u[:,t] )
            self.model.addConstr( self.u[:,t] <= u_max )

    def AddStateBounds(self, x_min, x_max):
        for t in range(self.T):
            self.model.addConstr( x_min <= self.x[:,t] )
            self.model.addConstr( self.x[:,t] <= x_max )

    def AddQuadraticCost(self, Q, R):
        raise NotImplementedError("Quadratic costs are not currently supported. Please use the DrakeMICPSolver for specifications with quadratic costs.")

    def AddRobustnessConstraint(self, rho_min=0.0):
        self.model.addConstr( self.rho >= rho_min )

    def Solve(self):
        self.model.optimize()

        if self.model.status == GRB.OPTIMAL:
            if self.verbose:
                print("\nOptimal Solution Found!\n")
            x = self.x.X
            u = self.u.X
            rho = self.rho.X[0]

            # Report optimal cost and robustness
            if self.verbose:
                print("Solve time: ", self.model.Runtime)
                print("Optimal robustness: ", rho)
                print("")
        else:
            if self.verbose:
                print(f"\nOptimization failed with status {self.model.status}.\n")
            x = None
            u = None
            rho = -np.inf

        return (x,u,rho,self.model.Runtime)

    def AddDynamicsConstraints(self):
        # Initial condition
        self.model.addConstr( self.x[:,0] == self.x0 )

        # Dynamics
        for t in range(self.T-1):
            self.model.addConstr(
                    self.x[:,t+1] == self.sys.A@self.x[:,t] + self.sys.B@self.u[:,t] )

            self.model.addConstr(
                    self.y[:,t] == self.sys.C@self.x[:,t] + self.sys.D@self.u[:,t] )

        self.model.addConstr(
                self.y[:,self.T-1] == self.sys.C@self.x[:,self.T-1] + self.sys.D@self.u[:,self.T-1] )

    def AddRobustnessCost(self):
        self.model.setObjective(1*self.rho, GRB.MAXIMIZE)

    def AddSTLConstraints(self):
        """
        Add the STL constraints

            (x,u) |= specification

        to the optimization problem, via the recursive introduction
        of binary variables for all subformulas in the specification.
        """
        # Recursively traverse the tree defined by the specification
        # to add binary variables and constraints that ensure that
        # rho is the robustness value
        z_spec = self.model.addMVar(1,vtype=GRB.BINARY)
        self.AddSubformulaConstraints(self.spec, z_spec, 0)
        self.model.addConstr( z_spec == 1 )

    def AddSubformulaConstraints(self, formula, z, t):
        """
        Given an STLFormula (formula) and a binary variable (z),
        add constraints to the optimization problem such that z
        takes value 1 only if the formula is satisfied (at time t).

        If the formula is a predicate, this constraint uses the "big-M"
        formulation

            A[x(t);u(t)] - b + (1-z)M >= 0,

        which enforces A[x;u] - b >= 0 if z=1, where (A,b) are the
        linear constraints associated with this predicate.

        If the formula is not a predicate, we recursively traverse the
        subformulas associated with this formula, adding new binary
        variables z_i for each subformula and constraining

            z <= z_i  for all i

        if the subformulas are combined with conjunction (i.e. all
        subformulas must hold), or otherwise constraining

            z <= sum(z_i)

        if the subformulas are combined with disjuction (at least one
        subformula must hold).
        """
        # We're at the bottom of the tree, so add the big-M constraints
        if isinstance(formula, LinearPredicate):
            # a.T*y - b + (1-z)*M >= rho
            self.model.addConstr( formula.a.T@self.y[:,t] - formula.b + (1-z)*self.M  >= self.rho )

            # Force z to be binary
            b = self.model.addMVar(1,vtype=GRB.BINARY)
            self.model.addConstr(z == b)
        
        elif isinstance(formula, NonlinearPredicate):
            raise TypeError("Mixed integer programming does not support nonlinear predicates")

        # We haven't reached the bottom of the tree, so keep adding
        # boolean constraints recursively
        else:
            if formula.combination_type == "and":
                for i, subformula in enumerate(formula.subformula_list):
                    z_sub = self.model.addMVar(1,vtype=GRB.CONTINUOUS)
                    t_sub = formula.timesteps[i]   # the timestep at which this formula
                                                   # should hold
                    self.AddSubformulaConstraints(subformula, z_sub, t+t_sub)
                    self.model.addConstr( z <= z_sub )

            else:  # combination_type == "or":
                z_subs = []
                for i, subformula in enumerate(formula.subformula_list):
                    z_sub = self.model.addMVar(1,vtype=GRB.CONTINUOUS)
                    z_subs.append(z_sub)
                    t_sub = formula.timesteps[i]
                    self.AddSubformulaConstraints(subformula, z_sub, t+t_sub)
                self.model.addConstr( z <= sum(z_subs) )




import numpy as np
import matplotlib.pyplot as plt

from .base import BenchmarkScenario
from .common import (inside_rectangle_formula,
                     outside_rectangle_formula,
                     make_rectangle_patch)
from ..systems import DoubleIntegrator

class RandomMultitarget(BenchmarkScenario):
    r"""
    A 2D mobile robot with double integrator dynamics must 
    navigate through a field of obstacles (grey, :math:`\mathcal{O}_i`)
    and reach at least one target of each color (:math:`\mathcal{T}_i^j`):

    .. math::

        \varphi = 
            \bigwedge_{i=1}^{N_c} \left( \bigvee_{j=1}^{N_t} F_{[0,T]} T_{i}^{j} \right) 
            \land G_{[0,T]} (\bigwedge_{k=1}^{N_o} \lnot O_k),

    :param num_obstacles:       number of obstacles, :math:`N_o`
    :param num_groups:          number of target groups/colors, :math:`N_c`
    :param targets_per_group:   number of targets in each group, :math:`N_t`
    :param T:                   time horizon of the specification
    :param seed:                (optional) seed for random generation of obstacle 
                                and target locations. Default is ``None``.
    """
    def __init__(self, num_obstacles, num_groups, targets_per_group, T, seed=None):
        self.T = T
        self.targets_per_group = targets_per_group

        # Set the seed for the random number generator (for reproducability)
        np.random.seed(seed)

        # Create the (randomly generated) set of obstacles
        self.obstacles = []
        for i in range(num_obstacles):
            x = np.random.uniform(0,9)  # keep within workspace
            y = np.random.uniform(0,9)
            self.obstacles.append((x,x+2,y,y+2))

        # Create the (randomly generated) set of targets
        self.targets = []
        for i in range(num_groups):
            target_group = []
            for j in range(targets_per_group):
                x = np.random.uniform(0,9)
                y = np.random.uniform(0,9)
                target_group.append((x,x+1,y,y+1))
            self.targets.append(target_group)

        self.T = T

    def GetSpecification(self):
        # Specify that we must avoid all obstacles
        obstacle_formulas = []
        for obs in self.obstacles:
            obstacle_formulas.append(outside_rectangle_formula(obs, 0, 1, 6))
        obstacle_avoidance = obstacle_formulas[0]
        for i in range(1, len(obstacle_formulas)):
            obstacle_avoidance = obstacle_avoidance & obstacle_formulas[i]

        # Specify that for each target group, we need to visit at least one
        # of the targets in that group
        target_group_formulas = []
        for target_group in self.targets:
            group_formulas = []
            for target in target_group:
                group_formulas.append(inside_rectangle_formula(target, 0, 1, 6))
            reach_target_group = group_formulas[0]
            for i in range(1, self.targets_per_group):
                reach_target_group = reach_target_group | group_formulas[i]
            target_group_formulas.append(reach_target_group)

        # Put all of the constraints together in one specification
        specification = obstacle_avoidance.always(0, self.T)
        for reach_target_group in target_group_formulas:
            specification = specification & reach_target_group.eventually(0, self.T)

        return specification

    def GetSystem(self):
        return DoubleIntegrator(2)

    def add_to_plot(self, ax):
        # Add red rectangles for the obstacles
        for obstacle in self.obstacles:
            ax.add_patch(make_rectangle_patch(*obstacle, color='k', alpha=0.5, zorder=-1))

        # Use the color cycle to choose the colors of each target group
        # (note that this won't work for more than 10 target groups)
        colors = plt.cm.tab10.colors
        for i, target_group in enumerate(self.targets):
            color = colors[i]
            for target in target_group:
                ax.add_patch(make_rectangle_patch(*target, color=color, alpha=0.7, zorder=-1))

        # set the field of view
        ax.set_xlim((0,10))
        ax.set_ylim((0,10))
        ax.set_aspect('equal')



##
#
# Common helper functions for defining STL specifications.
#
##

import numpy as np
from stlpy.STL import LinearPredicate, NonlinearPredicate
from matplotlib.patches import Rectangle, Circle

def inside_circle_formula(center, radius, y1_index, y2_index, d, name=None):
    """
    Create an STL formula representing being inside a
    circle with the given center and radius.

    :param center:      Tuple ``(y1, y2)`` specifying the center of the
                        circle.
    :param radius:      Radius of the circle
    :param y1_index:    index of the first (``y1``) dimension
    :param y2_index:    index of the second (``y2``) dimension
    :param d:           dimension of the overall signal
    :param name:        (optional) string describing this formula

    :return inside_circle:   A ``NonlinearPredicate`` specifying being inside the
                             circle at time zero.
    """
    # Define the predicate function g(y) >= 0
    def g(y):
        y1 = y[y1_index]
        y2 = y[y2_index]
        return radius**2 - (y1-center[0])**2 - (y2-center[1])**2

    return NonlinearPredicate(g, d, name=name)


def inside_rectangle_formula(bounds, y1_index, y2_index, d, name=None):
    """
    Create an STL formula representing being inside a
    rectangle with the given bounds:

    ::

       y2_max   +-------------------+
                |                   |
                |                   |
                |                   |
       y2_min   +-------------------+
                y1_min              y1_max

    :param bounds:      Tuple ``(y1_min, y1_max, y2_min, y2_max)`` containing
                        the bounds of the rectangle. 
    :param y1_index:    index of the first (``y1``) dimension
    :param y2_index:    index of the second (``y2``) dimension
    :param d:           dimension of the overall signal
    :param name:        (optional) string describing this formula

    :return inside_rectangle:   An ``STLFormula`` specifying being inside the
                                rectangle at time zero.
    """
    assert y1_index < d , "index must be less than signal dimension"
    assert y2_index < d , "index must be less than signal dimension"

    # Unpack the bounds
    y1_min, y1_max, y2_min, y2_max = bounds

    # Create predicates a*y >= b for each side of the rectangle
    a1 = np.zeros((1,d)); a1[:,y1_index] = 1
    right = LinearPredicate(a1, y1_min)
    left = LinearPredicate(-a1, -y1_max)

    a2 = np.zeros((1,d)); a2[:,y2_index] = 1
    top = LinearPredicate(a2, y2_min)
    bottom = LinearPredicate(-a2, -y2_max)

    # Take the conjuction across all the sides
    inside_rectangle = right & left & top & bottom

    # set the names
    if name is not None:
        right.name = "right of " + name
        left.name = "left of " + name
        top.name = "top of " + name
        bottom.name = "bottom of " + name
        inside_rectangle.name = name

    return inside_rectangle

def outside_rectangle_formula(bounds, y1_index, y2_index, d, name=None):
    """
    Create an STL formula representing being outside a
    rectangle with the given bounds:

    ::

       y2_max   +-------------------+
                |                   |
                |                   |
                |                   |
       y2_min   +-------------------+
                y1_min              y1_max

    :param bounds:      Tuple ``(y1_min, y1_max, y2_min, y2_max)`` containing
                        the bounds of the rectangle. 
    :param y1_index:    index of the first (``y1``) dimension
    :param y2_index:    index of the second (``y2``) dimension
    :param d:           dimension of the overall signal
    :param name:        (optional) string describing this formula
    
    :return outside_rectangle:   An ``STLFormula`` specifying being outside the
                                 rectangle at time zero.
    """
    assert y1_index < d , "index must be less than signal dimension"
    assert y2_index < d , "index must be less than signal dimension"

    # Unpack the bounds
    y1_min, y1_max, y2_min, y2_max = bounds

    # Create predicates a*y >= b for each side of the rectangle
    a1 = np.zeros((1,d)); a1[:,y1_index] = 1
    right = LinearPredicate(a1, y1_max)
    left = LinearPredicate(-a1, -y1_min)

    a2 = np.zeros((1,d)); a2[:,y2_index] = 1
    top = LinearPredicate(a2, y2_max)
    bottom = LinearPredicate(-a2, -y2_min)

    # Take the disjuction across all the sides
    outside_rectangle = right | left | top | bottom

    # set the names
    if name is not None:
        right.name = "right of " + name
        left.name = "left of " + name
        top.name = "top of " + name
        bottom.name = "bottom of " + name
        outside_rectangle.name = name

    return outside_rectangle

def make_rectangle_patch(xmin, xmax, ymin, ymax, **kwargs):
    """
    Convienience function for making a ``matplotlib.patches.Rectangle`` 
    patch for visualizing a rectangle:

    ::

       ymax   +-------------------+
              |                   |
              |                   |
              |                   |
       ymin   +-------------------+
              xmin                xmax

    :param xmin:        horizontal lower bound of the rectangle.
    :param xmax:        horizontal upper bound of the rectangle.
    :param ymin:        vertical lower bound of the rectangle.
    :param ymax:        vertical upper bound of the rectangle.
    :param kwargs:    (optional) keyword arguments passed to
                        the ``Rectangle`` constructor.

    :return patch:  a ``matplotlib.patches.Rectangle`` patch.

    """
    x = xmin
    y = ymin
    width = xmax-x
    height = ymax-y

    return Rectangle((x,y), width, height, **kwargs)

def make_circle_patch(center, radius, **kwargs):
    """
    Convienience function for making a ``matplotlib.patches.Circle`` 
    patch for visualizing a circle with the given center and radius.

    :param center:  Tuple containing the center coordinates of the circle
    :param radius:  The circle's radius
    :param kwargs:  (optional) keyword arguments passed to
                        the ``Circle`` constructor.

    :return patch:  a ``matplotlib.patches.Circle`` patch.
    """
    return Circle(center, radius, **kwargs)



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(name='stlpy',
      version='0.2.0',
      description='A Python library for control from Signal Temporal Logic (STL) specifications',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://stlpy.readthedocs.io/en/latest/index.html',
      author='Vince Kurtz',
      author_email='vjkurtz@gmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires=['numpy', 'matplotlib', 'treelib'],
      python_requires='>=3.8',
      zip_safe=False)



from .base import BenchmarkScenario
from .common import (inside_rectangle_formula,
                     outside_rectangle_formula,
                     make_rectangle_patch)
from ..systems import DoubleIntegrator

class ReachAvoid(BenchmarkScenario):
    r"""
    A 2D mobile robot with double integrator dynamics must
    avoid an obstacle (:math:`\mathcal{O}`) before reaching a goal (:math:`\mathcal{G}`):

    .. math::

        \varphi = G_{[0,T]} \lnot \mathcal{O} \land F_{[0,T]} \mathcal{G}

    :param goal_bounds:      a tuple ``(xmin, xmax, ymin, ymax)`` defining a
                             rectangular goal region.
    :param obstacle_bounds:  a tuple ``(xmin, xmax, ymin, ymax)`` defining a
                             rectangular obstacle.
    :param T:                the time horizon for this scenario.
    """
    def __init__(self, goal_bounds, obstacle_bounds, T):
        self.goal_bounds = goal_bounds
        self.obstacle_bounds = obstacle_bounds
        self.T = T

    def GetSpecification(self):
        # Goal Reaching
        at_goal = inside_rectangle_formula(self.goal_bounds, 0, 1, 6)

        # Obstacle Avoidance
        not_at_obstacle = outside_rectangle_formula(self.obstacle_bounds, 0, 1, 6)

        # Put all of the constraints together in one specification
        spec = not_at_obstacle.always(0, self.T) & at_goal.eventually(0, self.T)

        return spec

    def GetSystem(self):
        sys = DoubleIntegrator(2)
        return sys

    def add_to_plot(self, ax):
        # Make and add rectangular patches
        obstacle = make_rectangle_patch(*self.obstacle_bounds, color='k', alpha=0.5)
        goal = make_rectangle_patch(*self.goal_bounds, color='green', alpha=0.5)
        ax.add_patch(obstacle)
        ax.add_patch(goal)

        # set the field of view
        ax.set_xlim((0,10))
        ax.set_ylim((0,10))
        ax.set_aspect('equal')
